import re
import traceback

from hopy.const import TokenType, Identifier, TOKEN_EXPRS


class Token:
    def __init__(self, value, ttype, identifier, start):
        self.value = value
        self.ttype = ttype
        self.identifier = identifier
        self.start = start

    def __repr__(self):
        return f'Token({self.value!r}, {self.ttype!r}, {self.identifier!r}, {self.start!r})'


class IllegalInputException(Exception):
    pass


class Tokenizer:
    def __init__(self, inp, token_exprs=TOKEN_EXPRS, jsonpath_parser=None):
        self.compiled_token_exprs = self._compile_token_exprs(token_exprs)
        self.inp = inp
        self.i = 0
        self.tokens = []
        self.context = []
        self.jsonpath_parser = jsonpath_parser or self._default_jsonpath_parser()

    @staticmethod
    def _default_jsonpath_parser():
        from jsonpath_ng import parse as jsonpath_parse
        return jsonpath_parse

    @staticmethod
    def _compile_token_exprs(token_exprs):
        return [(re.compile(pattern), identifier, ttype) for pattern, identifier, ttype in token_exprs]

    def _next_match(self):
        max_match = None

        for regex, identifier, ttype in self.compiled_token_exprs:
            regex_match = regex.match(self.inp, self.i)
            if regex_match and (max_match is None or regex_match.end(0) > max_match[0].end(0)):
                max_match = (regex_match, identifier, ttype)

        if not max_match:
            raise IllegalInputException(f"Illegal character '{self.inp[self.i]}' encountered at position {self.i}")

        return max_match

    def _create_token(self, match):
        regex_match, identifier, ttype = match
        if ttype != TokenType.SKIP:
            if identifier == Identifier.OPEN_BRACE:
                self.context.append(Identifier.OPEN_BRACE)
                self.tokens.append(Token(regex_match.group(0), ttype, identifier, self.i))
            elif identifier == Identifier.CLOSE_BRACE:
                if self.context and self.context[-1] != Identifier.OPEN_BRACE:
                    raise IllegalInputException(f"Unmatched parenthesis at {self.i}")
                else:
                    self.context.pop()
                    self.tokens.append(Token(regex_match.group(0), ttype, identifier, self.i))
            elif identifier == Identifier.NUMBER:
                value = float(regex_match.group(0))
                self.tokens.append(Token(value, ttype, identifier, self.i))
            elif identifier == Identifier.STRING:
                value = regex_match.group(0)[1:-1]
                self.tokens.append(Token(value, ttype, identifier, self.i))
            elif identifier == Identifier.BOOLEAN:
                value = regex_match.group(0) in ('true', "TRUE", "True")
                self.tokens.append(Token(value, ttype, identifier, self.i))
            elif identifier == Identifier.JSONPATH:
                jsonpath_str = regex_match.group(0)[1:-1]
                value = self.jsonpath_parser(jsonpath_str)
                self.tokens.append(Token(value, ttype, identifier, self.i))
            else:
                self.tokens.append(Token(regex_match.group(0), ttype, identifier, self.i))

        self.i = regex_match.end(0)

    def tokenize(self):
        while self.i < len(self.inp):
            match = self._next_match()
            self._create_token(match)

        return self.tokens


if __name__ == '__main__':
    while True:
        inp = input("> Enter an expression to lex:\n> ")
        try:
            print(Tokenizer(inp).tokenize())
        except Exception as e:
            traceback.print_exc()



import traceback

from hopy.const import TokenType
from hopy.lexer import Tokenizer
from hopy.parser import Parser, ActionTreeGenerator


class Evaluator:
    def __init__(self, rules, tokenizer=Tokenizer, parser=Parser, action_tree_generator=ActionTreeGenerator):
        tokens = tokenizer(rules).tokenize()
        ast = parser(tokens).parse()
        self._action_tree = action_tree_generator(ast).generate_actions()

    def _inorder_evaluate(self, act_node, rule_input):
        if act_node.ast_node.token.ttype == TokenType.LITERAL:
            return act_node.ast_node.token.value
        elif act_node.ast_node.token.ttype == TokenType.EXTERNAL:
            return act_node.action(act_node.ast_node.token.value, rule_input)
        else:
            return act_node.action(*[self._inorder_evaluate(child, rule_input) for child in act_node.children])

    def evaluate(self, rule_input):
        return self._inorder_evaluate(self._action_tree, rule_input)


if __name__ == '__main__':
    while True:
        inp = input("> Enter an expression to lex:\n> ")
        try:
            evaluator = Evaluator(inp)
            print(evaluator.evaluate({"a": "1", "b": 2}))
        except Exception as e:
            traceback.print_exc()



class Pattern:
    BLANK_SPACE = r'[ \n\r\t]+'
    COMMA = r','
    OPEN_BRACE = r'\('
    CLOSE_BRACE = r'\)'
    EQUALS = r'=='
    NOT_EQUALS = r'!='
    GREATER_EQUALS = r'>='
    GREATER = r'>'
    LESSER_EQUALS = r'<='
    LESSER = r'<'
    AND = r'&&'
    OR = r'\|\|'
    NOT = r'!'
    NUMBER = r'[+-]?(\d*\.)?\d+'
    BOOLEAN = r'true|false|TRUE|FALSE|True|False'
    STRING = r"\"([^\"\\\n]|\\.|\\\n)*\"|'([^'\\\n]|\\.|\\\n)*'"
    JSONPATH = r'`([^`\\\n]|\\.|\\\n)*`'
    FUNCTION = r'f\.[a-z]\w+'


class Identifier:
    BLANK_SPACE = 'BLANK_SPACE'
    COMMA = 'COMMA'
    OPEN_BRACE = 'OPEN_BRACE'
    CLOSE_BRACE = 'CLOSE_BRACE'
    EQUALS = 'EQUALS'
    NOT_EQUALS = 'NOT_EQUALS'
    GREATER_EQUALS = 'GREATER_EQUALS'
    GREATER = 'GREATER'
    LESSER_EQUALS = 'LESSER_EQUALS'
    LESSER = 'LESSER'
    AND = 'AND'
    OR = 'OR'
    NOT = 'NOT'
    NUMBER = 'NUMBER'
    BOOLEAN = 'BOOLEAN'
    STRING = 'STRING'
    JSONPATH = 'JSONPATH'
    FUNCTION = 'FUNCTION'


class TokenType:
    EXTERNAL = 'EXTERNAL'
    LITERAL = 'LITERAL'
    OPERATOR = 'OPERATOR'
    SEPARATOR = 'SEPARATOR'
    SKIP = 'SKIP'


class OperatorType:
    UNARY = 'UNARY'
    BINARY = 'BINARY'
    NARY = 'NARY'


TOKEN_EXPRS = (
    (Pattern.BLANK_SPACE, Identifier.BLANK_SPACE, TokenType.SKIP),
    (Pattern.COMMA, Identifier.COMMA, TokenType.SEPARATOR),
    (Pattern.OPEN_BRACE, Identifier.OPEN_BRACE, TokenType.SEPARATOR),
    (Pattern.CLOSE_BRACE, Identifier.CLOSE_BRACE, TokenType.SEPARATOR),
    (Pattern.EQUALS, Identifier.EQUALS, TokenType.OPERATOR),
    (Pattern.NOT_EQUALS, Identifier.NOT_EQUALS, TokenType.OPERATOR),
    (Pattern.GREATER_EQUALS, Identifier.GREATER_EQUALS, TokenType.OPERATOR),
    (Pattern.GREATER, Identifier.GREATER, TokenType.OPERATOR),
    (Pattern.LESSER_EQUALS, Identifier.LESSER_EQUALS, TokenType.OPERATOR),
    (Pattern.LESSER, Identifier.LESSER, TokenType.OPERATOR),
    (Pattern.AND, Identifier.AND, TokenType.OPERATOR),
    (Pattern.OR, Identifier.OR, TokenType.OPERATOR),
    (Pattern.NOT, Identifier.NOT, TokenType.OPERATOR),
    (Pattern.NUMBER, Identifier.NUMBER, TokenType.LITERAL),
    (Pattern.BOOLEAN, Identifier.BOOLEAN, TokenType.LITERAL),
    (Pattern.STRING, Identifier.STRING, TokenType.LITERAL),
    (Pattern.JSONPATH, Identifier.JSONPATH, TokenType.EXTERNAL),
    (Pattern.FUNCTION, Identifier.FUNCTION, TokenType.OPERATOR),
)

OP_TYPE_MAPPING = {
    Identifier.FUNCTION: OperatorType.NARY,
    Identifier.NOT: OperatorType.UNARY,
    Identifier.EQUALS: OperatorType.BINARY,
    Identifier.NOT_EQUALS: OperatorType.BINARY,
    Identifier.GREATER_EQUALS: OperatorType.BINARY,
    Identifier.GREATER: OperatorType.BINARY,
    Identifier.LESSER_EQUALS: OperatorType.BINARY,
    Identifier.LESSER: OperatorType.BINARY,
    Identifier.AND: OperatorType.BINARY,
    Identifier.OR: OperatorType.BINARY,
}

OP_PRECEDENCE = {
    Identifier.EQUALS: 2,
    Identifier.NOT_EQUALS: 2,
    Identifier.GREATER_EQUALS: 2,
    Identifier.GREATER: 2,
    Identifier.LESSER_EQUALS: 2,
    Identifier.LESSER: 2,
    Identifier.AND: 3,
    Identifier.OR: 3,
    Identifier.NOT: 1,
    Identifier.FUNCTION: 0,
}



import traceback

from hopy.const import TOKEN_EXPRS, TokenType, Identifier, OP_TYPE_MAPPING, OperatorType, OP_PRECEDENCE
from hopy.lexer import Tokenizer


class ASTNode:
    def __init__(self, token, children=None):
        if children is None:
            children = []
        self.token = token
        self.children = children

    def __repr__(self, level=0):
        tabs = '\t' * level
        child_repr = "".join([child.__repr__(level + 1) for child in self.children])
        return f'{tabs}- {self.token.value}\n{child_repr}'


class ActionTreeNode:
    def __init__(self, ast_node, action, children=None, variable=False):
        if children is None:
            children = []
        self.ast_node = ast_node
        self.action = action
        self.variable = variable
        self.children = children


class Parser:
    def __init__(self, tokens):
        self.tokens = tokens
        self.op_stack = []
        self.nodes = []

    @staticmethod
    def _precedent(token1, token2):
        return OP_PRECEDENCE.get(token1.identifier, 99) <= OP_PRECEDENCE.get(token2.identifier, 99)

    def _backtrack(self):
        last_op = self.op_stack.pop()
        if OP_TYPE_MAPPING[last_op.identifier] == OperatorType.UNARY:
            operands = [self.nodes.pop()]
            self.nodes.append(ASTNode(last_op, operands))
        elif OP_TYPE_MAPPING[last_op.identifier] == OperatorType.BINARY:
            operand2 = self.nodes.pop()
            operand1 = self.nodes.pop()
            operands = [operand1, operand2]
            self.nodes.append(ASTNode(last_op, operands))
        else:
            raise Exception("Invalid operator encountered on stack")

    def parse(self):
        for token in self.tokens:
            if token.ttype == TokenType.OPERATOR:
                while len(self.op_stack) != 0 and self._precedent(self.op_stack[-1], token):
                    self._backtrack()
                self.op_stack.append(token)
            elif token.ttype == TokenType.SEPARATOR:
                if token.identifier == Identifier.OPEN_BRACE:
                    self.op_stack.append(token)
                elif token.identifier == Identifier.COMMA:
                    while len(self.op_stack) != 0 and self.op_stack[-1].ttype != TokenType.SEPARATOR:
                        self._backtrack()
                    self.op_stack.append(token)
                elif token.identifier == Identifier.CLOSE_BRACE:
                    while len(self.op_stack) != 0 and self.op_stack[-1].ttype != TokenType.SEPARATOR:
                        self._backtrack()

                    operands = []
                    while len(self.op_stack) != 0 and self.op_stack[-1].identifier != Identifier.OPEN_BRACE:
                        last_op = self.op_stack.pop()
                        if last_op.identifier == Identifier.COMMA:
                            operands.append(self.nodes.pop())
                        else:
                            raise Exception("Invalid operator found on stack")

                    if len(self.op_stack) == 0 or self.op_stack[-1].identifier != Identifier.OPEN_BRACE:
                        raise Exception("Invalid closing parenthesis")

                    else:
                        operands.append(self.nodes.pop())
                        self.op_stack.pop()

                    operands.reverse()
                    if len(self.op_stack) != 0 and OP_TYPE_MAPPING[self.op_stack[-1].identifier] == OperatorType.NARY:
                        self.nodes.append(ASTNode(self.op_stack.pop(), operands))
                    else:
                        self.nodes.append(operands[0])

            else:
                self.nodes.append(ASTNode(token))

        while len(self.op_stack) != 0:
            self._backtrack()

        return self.nodes[0]


class ActionTreeGenerator:
    def __init__(self, ast):
        self._ast = ast
        self._ast_act_map = {}

    @staticmethod
    def _get_equals_action(token):
        return (lambda a, b: a == b), False

    @staticmethod
    def _get_not_equals_action(token):
        return (lambda a, b: a != b), False

    @staticmethod
    def _get_greater_equals_action(token):
        return (lambda a, b: a >= b), False

    @staticmethod
    def _get_greater_action(token):
        return (lambda a, b: a > b), False

    @staticmethod
    def _get_lesser_equals_action(token):
        return (lambda a, b: a <= b), False

    @staticmethod
    def _get_lesser_action(token):
        return (lambda a, b: a < b), False

    @staticmethod
    def _get_and_action(token):
        return (lambda a, b: a and b), False

    @staticmethod
    def _get_or_action(token):
        return (lambda a, b: a or b), False

    @staticmethod
    def _get_not_action(token):
        return (lambda a: not a), False

    @staticmethod
    def _get_jsonpath_action(token):
        return (lambda x, y: x.find(y)[0].value), True

    @staticmethod
    def _get_function_action(token):
        if token.value == 'f.abs':
            return abs, False
        elif token.value == 'f.max':
            return max, False
        elif token.value == 'f.min':
            return min, False
        else:
            raise Exception("Invalid function found")

    def _get_action(self, token):
        if token.ttype in (TokenType.OPERATOR, TokenType.EXTERNAL):
            return getattr(self, f'_get_{token.identifier.lower()}_action')(token)
        elif token.ttype == TokenType.LITERAL:
            return token.value, False
        else:
            raise Exception("Invalid token type")

    def _get_act_node_shallow(self, ast_node):
        action, variable = self._get_action(ast_node.token)
        return ActionTreeNode(ast_node, action, variable=variable)

    def _inorder(self, ast_node):
        act_node = self._get_act_node_shallow(ast_node)
        for child in ast_node.children:
            act_node.children.append(self._inorder(child))

        return act_node

    def generate_actions(self):
        return self._inorder(self._ast)


if __name__ == '__main__':
    while True:
        inp = input("> Enter an expression to lex:\n> ")
        try:
            tokens = Tokenizer(TOKEN_EXPRS, inp).tokenize()
            print(tokens)
            ast = Parser(tokens).parse()
            print(ast)
        except Exception as e:
            traceback.print_exc()
